import functools

from scipy.special import gamma
from scipy.stats import binned_statistic, iqr
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import iqr
import torch
from scipy.optimize import minimize

from ..utils.initialize import isophotes
from ..utils.parametric_profiles import (
    sersic_torch,
    sersic_np,
    gaussian_torch,
    gaussian_np,
    exponential_torch,
    exponential_np,
    spline_torch,
    moffat_torch,
    moffat_np,
    nuker_torch,
    nuker_np,
)
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.conversions.coordinates import Rotate_Cartesian
from ..utils.conversions.functions import sersic_I0_to_flux_np, sersic_flux_to_I0_torch
from ..image import (
    Image_List,
    Target_Image,
    Model_Image_List,
    Target_Image_List,
    Window_List,
)
from ..param import Param_Unlock, Param_SoftLimits
from .. import AP_config

# Target Selector Decorator
######################################################################
def select_target(func):
    @functools.wraps(func)
    def targeted(self, target=None, **kwargs):
        if target is None:
            send_target = self.target
        elif isinstance(target, Target_Image_List) and not isinstance(
            self.target, Image_List
        ):
            for sub_target in target:
                if sub_target.identity == self.target.identity:
                    send_target = sub_target
                    break
            else:
                raise RuntimeError(
                    "{self.name} could not find matching target to initialize with"
                )
        else:
            send_target = target
        return func(self, target=send_target, **kwargs)

    return targeted


def select_sample(func):
    @functools.wraps(func)
    def targeted(self, image=None, **kwargs):
        if isinstance(image, Model_Image_List) and not isinstance(
            self.target, Image_List
        ):
            for i, sub_image in enumerate(image):
                if sub_image.target_identity == self.target.identity:
                    send_image = sub_image
                    if "window" in kwargs and isinstance(kwargs["window"], Window_List):
                        kwargs["window"] = kwargs["window"].window_list[i]
                    break
            else:
                raise RuntimeError(
                    "{self.name} could not find matching image to sample with"
                )
        else:
            send_image = image
        return func(self, image=send_image, **kwargs)

    return targeted


# General parametric
######################################################################
@torch.no_grad()
@ignore_numpy_warnings
def parametric_initialize(
    model, parameters, target, prof_func, params, x0_func, force_uncertainty=None
):
    if all(list(parameters[param].value is not None for param in params)):
        return
    # Get the sub-image area corresponding to the model image
    target_area = target[model.window]
    target_dat = target_area.data.detach().cpu().numpy()
    if target_area.has_mask:
        mask = target_area.mask.detach().cpu().numpy()
        target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
    edge = np.concatenate(
        (
            target_dat[:, 0],
            target_dat[:, -1],
            target_dat[0, :],
            target_dat[-1, :],
        )
    )
    edge_average = np.median(edge)
    edge_scatter = iqr(edge, rng=(16, 84)) / 2
    # Convert center coordinates to target area array indices
    icenter = target_area.plane_to_pixel(parameters["center"].value)
        
    # Collect isophotes for 1D fit
    iso_info = isophotes(
        target_dat - edge_average,
        (icenter[1].item(), icenter[0].item()),
        threshold=3 * edge_scatter,
        pa=(parameters["PA"].value - target.north).detach().cpu().item()
        if "PA" in parameters
        else 0.0,
        q=parameters["q"].value.detach().cpu().item() if "q" in parameters else 1.0,
        n_isophotes=15,
    )
    R = np.array(list(iso["R"] for iso in iso_info)) * target.pixel_length.item()
    flux = np.array(list(iso["flux"] for iso in iso_info)) / target.pixel_area.item()
    # Correct the flux if values are negative, so fit can be done in log space
    if np.sum(flux < 0) > 0:
        AP_config.ap_logger.debug("fixing flux")
        flux -= np.min(flux) - np.abs(np.min(flux) * 0.1)
    flux = np.log10(flux)

    x0 = list(x0_func(model, R, flux))
    for i, param in enumerate(params):
        x0[i] = (
            x0[i] if parameters[param].value is None else parameters[param].value.item()
        )

    def optim(x, r, f):
        residual = (f - np.log10(prof_func(r, *x))) ** 2
        N = np.argsort(residual)
        return np.mean(residual[:-3])

    res = minimize(optim, x0=x0, args=(R, flux), method="Nelder-Mead")
    if not res.success and AP_config.ap_verbose >= 2:
        AP_config.ap_logger.warn(
            f"initialization fit not successful for {model.name}, falling back to defaults"
        )

    if force_uncertainty is None:
        reses = []
        for i in range(10):
            N = np.random.randint(0, len(R), len(R))
            reses.append(
                minimize(optim, x0=x0, args=(R[N], flux[N]), method="Nelder-Mead")
            )
    for param, resx, x0x in zip(params, res.x, x0):
        with Param_Unlock(parameters[param]), Param_SoftLimits(parameters[param]):
            if parameters[param].value is None:
                parameters[param].value = resx if res.success else x0x
            if force_uncertainty is None and parameters[param].uncertainty is None:
                parameters[param].uncertainty = np.std(list(subres.x[params.index(param)] for subres in reses))
            elif force_uncertainty is not None:
                parameters[param].uncertainty = force_uncertainty[params.index(param)]


@torch.no_grad()
@ignore_numpy_warnings
def parametric_segment_initialize(
    model=None,
    parameters=None,
    target=None,
    prof_func=None,
    params=None,
    x0_func=None,
    segments=None,
    force_uncertainty=None,
):
    if all(list(model[param].value is not None for param in params)):
        return
    # Get the sub-image area corresponding to the model image
    target_area = target[model.window]
    target_dat = target_area.data.detach().cpu().numpy()
    if target_area.has_mask:
        mask = target_area.mask.detach().cpu().numpy()
        target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
    edge = np.concatenate(
        (
            target_dat[:, 0],
            target_dat[:, -1],
            target_dat[0, :],
            target_dat[-1, :],
        )
    )
    edge_average = np.median(edge)
    edge_scatter = iqr(edge, rng=(16, 84)) / 2
    # Convert center coordinates to target area array indices
    icenter = target_area.plane_to_pixel(model["center"].value)

    iso_info = isophotes(
        target_dat - edge_average,
        (icenter[1].item(), icenter[0].item()),
        threshold=3 * edge_scatter,
        pa=(model["PA"].value - target.north).item() if "PA" in model else 0.0,
        q=model["q"].value.item() if "q" in model else 1.0,
        n_isophotes=15,
        more=True,
    )
    R = np.array(list(iso["R"] for iso in iso_info)) * target.pixel_length.item()
    was_none = list(False for i in range(len(params)))
    val = {}
    unc = {}
    for i, p in enumerate(params):
        if model[p].value is None:
            was_none[i] = True
            val[p] = np.zeros(segments)
            unc[p] = np.zeros(segments)
    for r in range(segments):
        flux = []
        for iso in iso_info:
            modangles = (
                iso["angles"]
                - (
                    (model["PA"].value - target.north).detach().cpu().item()
                    + r * np.pi / segments
                )
            ) % np.pi
            flux.append(
                np.median(
                    iso["isovals"][
                        np.logical_or(
                            modangles < (0.5 * np.pi / segments),
                            modangles >= (np.pi * (1 - 0.5 / segments)),
                        )
                    ]
                )
            )
        flux = np.array(flux) / target.pixel_area.item()
        if np.sum(flux < 0) >= 1:
            flux -= np.min(flux) - np.abs(np.min(flux) * 0.1)
        flux = np.log10(flux)

        x0 = list(x0_func(model, R, flux))
        for i, param in enumerate(params):
            x0[i] = (
                x0[i] if was_none[i] else model[param].value.detach().cpu().numpy()[r]
            )
        res = minimize(
            lambda x: np.mean((flux - np.log10(prof_func(R, *x))) ** 2),
            x0=x0,
            method="Nelder-Mead",
        )
        if force_uncertainty is None:
            reses = []
            for i in range(10):
                N = np.random.randint(0, len(R), len(R))
                reses.append(
                    minimize(
                        lambda x: np.mean((flux - np.log10(prof_func(R, *x))) ** 2),
                        x0=x0,
                        method="Nelder-Mead",
                    )
                )
        for i, param in enumerate(params):
            if was_none[i]:
                val[param][r] = res.x[i] if res.success else x0[i]
                if force_uncertainty is None and model[param].uncertainty is None:
                    unc[r] = np.std(list(subres.x[params.index(param)] for subres in reses))
                elif force_uncertainty is not None:
                    unc[r] = force_uncertainty[params.index(param)][r]
                
            with Param_Unlock(model[param]), Param_SoftLimits(model[param]):
                model[param].value = val[param]
                model[param].uncertainty = unc[param]


# Evaluate_Model
######################################################################
@default_internal
def radial_evaluate_model(self, X=None, Y=None, image=None, parameters=None):
    if X is None:
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - parameters["center"].value[..., None, None]
    return self.radial_model(
        self.radius_metric(X, Y, image=image, parameters=parameters),
        image=image,
        parameters=parameters,
    )

                
# Exponential
######################################################################
@default_internal
def exponential_radial_model(self, R, image=None, parameters=None):
    return exponential_torch(
        R,
        parameters["Re"].value,
        image.pixel_area * 10 ** parameters["Ie"].value,
    )


@default_internal
def exponential_iradial_model(self, i, R, image=None, parameters=None):
    return exponential_torch(
        R,
        parameters["Re"].value[i],
        image.pixel_area * 10 ** parameters["Ie"].value[i],
    )


# Sersic
######################################################################
@default_internal
def sersic_radial_model(self, R, image=None, parameters=None):
    return sersic_torch(
        R,
        parameters["n"].value,
        parameters["Re"].value,
        image.pixel_area * 10 ** parameters["Ie"].value,
    )


@default_internal
def sersic_iradial_model(self, i, R, image=None, parameters=None):
    return sersic_torch(
        R,
        parameters["n"].value[i],
        parameters["Re"].value[i],
        image.pixel_area * 10 ** parameters["Ie"].value[i],
    )


# Moffat
######################################################################
@default_internal
def moffat_radial_model(self, R, image=None, parameters=None):
    return moffat_torch(
        R,
        parameters["n"].value,
        parameters["Rd"].value,
        image.pixel_area * 10 ** parameters["I0"].value,
    )


@default_internal
def moffat_iradial_model(self, i, R, image=None, parameters=None):
    return moffat_torch(
        R,
        parameters["n"].value[i],
        parameters["Rd"].value[i],
        image.pixel_area * 10 ** parameters["I0"].value[i],
    )


# Nuker Profile
######################################################################
@default_internal
def nuker_radial_model(self, R, image=None, parameters=None):
    return nuker_torch(
        R,
        parameters["Rb"].value,
        image.pixel_area * 10 ** parameters["Ib"].value,
        parameters["alpha"].value,
        parameters["beta"].value,
        parameters["gamma"].value,
    )


@default_internal
def nuker_iradial_model(self, i, R, image=None, parameters=None):
    return nuker_torch(
        R,
        parameters["Rb"].value[i],
        image.pixel_area * 10 ** parameters["Ib"].value[i],
        parameters["alpha"].value[i],
        parameters["beta"].value[i],
        parameters["gamma"].value[i],
    )


# Gaussian
######################################################################
@default_internal
def gaussian_radial_model(self, R, image=None, parameters=None):
    return gaussian_torch(
        R,
        parameters["sigma"].value,
        image.pixel_area * 10 ** parameters["flux"].value,
    )


@default_internal
def gaussian_iradial_model(self, i, R, image=None, parameters=None):
    return gaussian_torch(
        R,
        parameters["sigma"].value[i],
        image.pixel_area * 10 ** parameters["flux"].value[i],
    )


# Spline
######################################################################
@torch.no_grad()
@ignore_numpy_warnings
@select_target
@default_internal
def spline_initialize(self, target=None, parameters=None, **kwargs):
    super(self.__class__, self).initialize(target=target, parameters=parameters)

    if parameters["I(R)"].value is not None and parameters["I(R)"].prof is not None:
        return

    # Create the I(R) profile radii if needed
    if parameters["I(R)"].prof is None:
        new_prof = [0, 2 * target.pixel_length]
        while new_prof[-1] < torch.max(self.window.shape / 2):
            new_prof.append(
                new_prof[-1] + torch.max(2 * target.pixel_length, new_prof[-1] * 0.2)
            )
        new_prof.pop()
        new_prof.pop()
        new_prof.append(torch.sqrt(torch.sum((self.window.shape / 2) ** 2)))
        parameters["I(R)"].prof = new_prof

    profR = parameters["I(R)"].prof.detach().cpu().numpy()
    target_area = target[self.window]
    target_dat = target_area.data.detach().cpu().numpy()
    if target_area.has_mask:
        mask = target_area.mask.detach().cpu().numpy()
        target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
    Coords = target_area.get_coordinate_meshgrid()
    X, Y = Coords - parameters["center"].value[..., None, None]
    X, Y = self.transform_coordinates(X, Y, target, parameters)
    R = self.radius_metric(X, Y, target, parameters).detach().cpu().numpy()
    rad_bins = [profR[0]] + list((profR[:-1] + profR[1:]) / 2) + [profR[-1] * 100]
    raveldat = target_dat.ravel()

    I = (
        binned_statistic(R.ravel(), raveldat, statistic="median", bins=rad_bins)[0]
    ) / target.pixel_area.item()
    N = np.isfinite(I)
    if not np.all(N):
        I[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], I[N])
    I = np.abs(I)
    if I[-1] >= I[-2]:
        I[-1] = I[-2] / 2
    S = binned_statistic(
        R.ravel(), raveldat, statistic=lambda d: iqr(d, rng=[16, 84]) / 2, bins=rad_bins
    )[0]
    N = np.isfinite(S)
    if not np.all(N):
        S[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], S[N])
    with Param_Unlock(parameters["I(R)"]), Param_SoftLimits(parameters["I(R)"]):
        parameters["I(R)"].value = np.log10(I)
        parameters["I(R)"].uncertainty = S / (I * np.log(10))


@torch.no_grad()
@ignore_numpy_warnings
@select_target
@default_internal
def spline_segment_initialize(
    self, target=None, parameters=None, segments=1, symmetric=True, **kwargs
):
    super(self.__class__, self).initialize(target=target, parameters=parameters)

    if parameters["I(R)"].value is not None and parameters["I(R)"].prof is not None:
        return

    # Create the I(R) profile radii if needed
    if parameters["I(R)"].prof is None:
        new_prof = [0, 2 * target.pixel_length]
        while new_prof[-1] < torch.max(self.window.shape / 2):
            new_prof.append(
                new_prof[-1] + torch.max(2 * target.pixel_length, new_prof[-1] * 0.2)
            )
        new_prof.pop()
        new_prof.pop()
        new_prof.append(torch.sqrt(torch.sum((self.window.shape / 2) ** 2)))
        parameters["I(R)"].prof = new_prof

    profR = parameters["I(R)"].prof.detach().cpu().numpy()
    target_area = target[self.window]
    target_dat = target_area.data.detach().cpu().numpy()
    if target_area.has_mask:
        mask = target_area.mask.detach().cpu().numpy()
        target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
    Coords = target_area.get_coordinate_meshgrid()
    X, Y = Coords - parameters["center"].value[..., None, None]
    X, Y = self.transform_coordinates(X, Y, target, parameters)
    R = self.radius_metric(X, Y, target, parameters).detach().cpu().numpy()
    T = self.angular_metric(X, Y, target, parameters).detach().cpu().numpy()
    rad_bins = [profR[0]] + list((profR[:-1] + profR[1:]) / 2) + [profR[-1] * 100]
    raveldat = target_dat.ravel()
    val = np.zeros((segments, len(parameters["I(R)"].prof)))
    unc = np.zeros((segments, len(parameters["I(R)"].prof)))
    for s in range(segments):
        if segments % 2 == 0 and symmetric:
            angles = (T - (s * np.pi / segments)) % np.pi
            TCHOOSE = np.logical_or(
                angles < (np.pi / segments), angles >= (np.pi * (1 - 1 / segments))
            )
        elif segments % 2 == 1 and symmetric:
            angles = (T - (s * np.pi / segments)) % (2 * np.pi)
            TCHOOSE = np.logical_or(
                angles < (np.pi / segments), angles >= (np.pi * (2 - 1 / segments))
            )
            angles = (T - (np.pi + r * np.pi / segments)) % (2 * np.pi)
            TCHOOSE = np.logical_or(
                TCHOOSE,
                np.logical_or(
                    angles < (np.pi / segments), angles >= (np.pi * (2 - 1 / segments))
                ),
            )
        elif segments % 2 == 0 and not symmetric:
            angles = (T - (s * 2 * np.pi / segments)) % (2 * np.pi)
            TCHOOSE = torch.logical_or(
                angles < (2 * np.pi / segments),
                angles >= (2 * np.pi * (1 - 1 / segments)),
            )
        else:
            angles = (T - (s * 2 * np.pi / segments)) % (2 * np.pi)
            TCHOOSE = torch.logical_or(
                angles < (2 * np.pi / segments), angles >= (np.pi * (2 - 1 / segments))
            )
        TCHOOSE = TCHOOSE.ravel()
        I = (
            binned_statistic(
                R.ravel()[TCHOOSE], raveldat[TCHOOSE], statistic="median", bins=rad_bins
            )[0]
        ) / target.pixel_area.item()
        N = np.isfinite(I)
        if not np.all(N):
            I[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], I[N])
        S = binned_statistic(
            R.ravel(),
            raveldat,
            statistic=lambda d: iqr(d, rng=[16, 84]) / 2,
            bins=rad_bins,
        )[0]
        N = np.isfinite(S)
        if not np.all(N):
            S[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], S[N])
        val[s] = np.log10(np.abs(I))
        unc[s] = S / (np.abs(I) * np.log(10))
    with Param_Unlock(parameters["I(R)"]), Param_SoftLimits(parameters["I(R)"]):
        parameters["I(R)"].value = val
        parameters["I(R)"].uncertainty = unc


@default_internal
def spline_radial_model(self, R, image=None, parameters=None):
    return (
        spline_torch(
            R,
            parameters["I(R)"].prof,
            parameters["I(R)"].value,
            extend=self.extend_profile,
        )
        * image.pixel_area
    )


@default_internal
def spline_iradial_model(self, i, R, image=None, parameters=None):
    return (
        spline_torch(
            R,
            parameters["I(R)"].prof,
            parameters["I(R)"].value[i],
            extend=self.extend_profile,
        )
        * image.pixel_area
    )

# RelSpline
######################################################################
@torch.no_grad()
@ignore_numpy_warnings
@select_target
@default_internal
def relspline_initialize(self, target=None, parameters=None, **kwargs):
    super(self.__class__, self).initialize(target=target, parameters=parameters)

    target_area = target[self.window]
    target_dat = target_area.data.detach().cpu().numpy()
    if target_area.has_mask:
        mask = target_area.mask.detach().cpu().numpy()
        target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
    if parameters["I0"].value is None:
        center = target_area.plane_to_pixel(parameters["center"].value)
        flux = target_dat[center[1].int().item(), center[0].int().item()]
        with Param_Unlock(parameters["I0"]), Param_SoftLimits(parameters["I0"]):
            parameters["I0"].value = np.log10(np.abs(flux) / target_area.pixel_area.item())
            parameters["I0"].uncertainty = 0.01
        
    if parameters["dI(R)"].value is not None and parameters["dI(R)"].prof is not None:
        return

    # Create the I(R) profile radii if needed
    if parameters["dI(R)"].prof is None:
        new_prof = [2 * target.pixel_length]
        while new_prof[-1] < torch.max(self.window.shape / 2):
            new_prof.append(
                new_prof[-1] + torch.max(2 * target.pixel_length, new_prof[-1] * 0.2)
            )
        new_prof.pop()
        new_prof.pop()
        new_prof.append(torch.sqrt(torch.sum((self.window.shape / 2) ** 2)))
        parameters["dI(R)"].prof = new_prof

    profR = parameters["dI(R)"].prof.detach().cpu().numpy()
        
    Coords = target_area.get_coordinate_meshgrid()
    X, Y = Coords - parameters["center"].value[..., None, None]
    X, Y = self.transform_coordinates(X, Y, target, parameters)
    R = self.radius_metric(X, Y, target, parameters).detach().cpu().numpy()
    rad_bins = [profR[0]] + list((profR[:-1] + profR[1:]) / 2) + [profR[-1] * 100]
    raveldat = target_dat.ravel()

    I = (
        binned_statistic(R.ravel(), raveldat, statistic="median", bins=rad_bins)[0]
    ) / target.pixel_area.item()
    N = np.isfinite(I)
    if not np.all(N):
        I[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], I[N])
    if I[-1] >= I[-2]:
        I[-1] = I[-2] / 2
    S = binned_statistic(
        R.ravel(), raveldat, statistic=lambda d: iqr(d, rng=[16, 84]) / 2, bins=rad_bins
    )[0]
    N = np.isfinite(S)
    if not np.all(N):
        S[np.logical_not(N)] = np.interp(profR[np.logical_not(N)], profR[N], S[N])
    with Param_Unlock(parameters["dI(R)"]), Param_SoftLimits(parameters["dI(R)"]):
        parameters["dI(R)"].value = np.log10(np.abs(I)) - parameters["I0"].value.item()
        parameters["dI(R)"].uncertainty = S / (np.abs(I) * np.log(10))

@default_internal
def relspline_radial_model(self, R, image=None, parameters=None):
    return (
        spline_torch(
            R,
            torch.cat((torch.zeros_like(parameters["I0"].value).unsqueeze(-1),parameters["dI(R)"].prof)),
            torch.cat((parameters["I0"].value.unsqueeze(-1), parameters["I0"].value + parameters["dI(R)"].value)),
            extend=self.extend_profile,
        )
        * image.pixel_area
    )



import warnings
from typing import Optional, Union, List

import torch
from torch.nn.functional import pad

from .image_object import Image, Image_List
from .. import AP_config
from ..errors import SpecificationConflict, InvalidImage

__all__ = ["Jacobian_Image", "Jacobian_Image_List"]


######################################################################
class Jacobian_Image(Image):
    """Jacobian of a model evaluated in an image.

    Image object which represents the evaluation of a jacobian on an
    image. It takes the form of a 3D (Image x Nparameters)
    tensor. This object can be added other other Jacobian images to
    build up a full jacobian for a complex model.

    """

    def __init__(
        self,
        parameters: List[str],
        target_identity: str,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.target_identity = target_identity
        self.parameters = list(parameters)
        if len(self.parameters) != len(set(self.parameters)):
            raise SpecificationConflict("Every parameter should be unique upon jacobian creation")

    def flatten(self, attribute: str = "data"):
        return getattr(self, attribute).reshape((-1, len(self.parameters)))

    def copy(self, **kwargs):
        return super().copy(
            parameters=self.parameters, target_identity=self.target_identity, **kwargs
        )

    def get_state(self):
        state = super().get_state()
        state["target_identity"] = self.target_identity
        state["parameters"] = self.parameters
        return state

    def set_state(self, state):
        super().set_state(state)
        self.target_identity = state["target_identity"]
        self.parameters = state["parameters"]

    def get_fits_state(self):
        states = super().get_fits_state()
        for state in states:
            if state["HEADER"]["IMAGE"] == "PRIMARY":
                state["HEADER"]["TRGTID"] = self.target_identity
                state["HEADER"]["PARAMS"] = str(self.parameters)
        return states

    def set_fits_state(self, states):
        super().set_fits_state(states)
        for state in states:
            if state["HEADER"]["IMAGE"] == "PRIMARY":
                self.target_identity = state["HEADER"]["TRGTID"]
                self.parameters = eval(state["HEADER"]["params"])
        
    def __add__(self, other):
        raise NotImplementedError("Jacobian images cannot add like this, use +=")

    def __sub__(self, other):
        raise NotImplementedError("Jacobian images cannot subtract")

    def __isub__(self, other):
        raise NotImplementedError("Jacobian images cannot subtract")

    def __iadd__(self, other):
        if not isinstance(other, Jacobian_Image):
            raise InvalidImage("Jacobian images can only add with each other, not: type(other)")

        # exclude null jacobian images
        if other.data is None:
            return self
        if self.data is None:
            return other

        full_window = self.window | other.window

        self_indices = other.window.get_other_indices(self)
        other_indices = self.window.get_other_indices(other)
        for i, other_identity in enumerate(other.parameters):
            if other_identity in self.parameters:
                other_loc = self.parameters.index(other_identity)
            else:
                self.set_data(
                    torch.cat(
                        (
                            self.data,
                            torch.zeros(
                                self.data.shape[0],
                                self.data.shape[1],
                                1,
                                dtype=AP_config.ap_dtype,
                                device=AP_config.ap_device,
                            ),
                        ),
                        dim=2,
                    ),
                    require_shape=False,
                )
                self.parameters.append(other_identity)
                other_loc = -1
            self.data[self_indices[0], self_indices[1], other_loc] += other.data[
                other_indices[0], other_indices[1], i
            ]
        return self


######################################################################
class Jacobian_Image_List(Image_List, Jacobian_Image):
    """For joint modelling, represents Jacobians evaluated on a list of
    images.

    Stores jacobians evaluated on a number of image objects. Since
    jacobian images are aware of the target images they were evaluated
    on, it is possible to combine this object with other
    Jacobian_Image_List objects or even Jacobian_Image objects and
    everything will be sorted into the proper locations of the list,
    and image.

    """

    def __init__(self, image_list):
        super().__init__(image_list)

    def flatten(self, attribute="data"):
        if len(self.image_list) > 1:
            for image in self.image_list[1:]:
                if self.image_list[0].parameters != image.parameters:
                    raise SpecificationConflict("Jacobian image list sub-images track different parameters. Please initialize with all parameters that will be used.")
        return torch.cat(tuple(image.flatten(attribute) for image in self.image_list))

    def __add__(self, other):
        raise NotImplementedError("Jacobian images cannot add like this, use +=")

    def __sub__(self, other):
        raise NotImplementedError("Jacobian images cannot subtract")

    def __isub__(self, other):
        raise NotImplementedError("Jacobian images cannot subtract")

    def __iadd__(self, other):
        if isinstance(other, Jacobian_Image_List):
            for other_image in other.image_list:
                for self_image in self.image_list:
                    if other_image.target_identity == self_image.target_identity:
                        self_image += other_image
                        break
                else:
                    self.image_list.append(other_image)
        elif isinstance(other, Jacobian_Image):
            for self_image in self.image_list:
                if other.target_identity == self_image.target_identity:
                    self_image += other
                    break
            else:
                self.image_list.append(other_image)
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image += other_image
        return self



import requests

tutorials = [
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/GettingStarted.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/GroupModels.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/ModelZoo.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/JointModels.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/FittingMethods.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/CustomModels.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/AdvancedPSFModels.ipynb",
    "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/ConstrainedModels.ipynb",
]
for url in tutorials:
    try:
        R = requests.get(url)
        with open(url[url.rfind("/") + 1 :], "w") as f:
            f.write(R.text)
    except:
        print(
            f"WARNING: couldn't find tutorial: {url[url.rfind('/')+1:]} check internet conection"
        )



import numpy as np
import torch
from ... import AP_config


def boundaries(val, limits):
    """val in limits expanded to range -inf to inf"""
    tval = torch.as_tensor(val, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    
    if limits[0] is None:
        return tval - 1.0 / (tval - limits[1])
    elif limits[1] is None:
        return tval - 1.0 / (tval - limits[0])
    return torch.tan((tval - limits[0]) * np.pi / (limits[1] - limits[0]) - np.pi / 2)


def inv_boundaries(val, limits):
    """val in range -inf to inf compressed to within the limits"""
    tval = torch.as_tensor(val, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    
    if limits[0] is None:
        return (tval + limits[1] - torch.sqrt(torch.pow(tval - limits[1], 2) + 4)) * 0.5
    elif limits[1] is None:
        return (tval + limits[0] + torch.sqrt(torch.pow(tval - limits[0], 2) + 4)) * 0.5
    return (torch.arctan(tval) + np.pi / 2) * (limits[1] - limits[0]) / np.pi + limits[0]


def d_boundaries_dval(val, limits):
    """derivative of: val in limits expanded to range -inf to inf"""
    tval = torch.as_tensor(val, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    if limits[0] is None:
        return 1.0 + 1.0 / (tval - limits[1]) ** 2
    elif limits[1] is None:
        return 1.0 - 1.0 / (tval - limits[0]) ** 2
    return (np.pi / (limits[1] - limits[0])) / torch.cos(
        (tval - limits[0]) * np.pi / (limits[1] - limits[0]) - np.pi / 2
    ) ** 2


def d_inv_boundaries_dval(val, limits):
    """derivative of: val in range -inf to inf compressed to within the limits"""
    tval = torch.as_tensor(val, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    if limits[0] is None:
        return 0.5 - 0.5 * (tval - limits[1]) / torch.sqrt(
            torch.pow(tval - limits[1], 2) + 4
        )
    elif limits[1] is None:
        return 0.5 + 0.5 * (tval - limits[0]) / torch.sqrt(
            torch.pow(tval - limits[0], 2) + 4
        )
    return (limits[1] - limits[0]) / (np.pi * (tval ** 2 + 1))


def cyclic_boundaries(val, limits):
    """Applies cyclic boundary conditions to the input value."""
    tval = torch.as_tensor(val, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    
    return limits[0] + ((tval - limits[0]) % (limits[1] - limits[0]))


def cyclic_difference_torch(val1, val2, period):
    """Applies the difference operation between two values with cyclic
    boundary conditions.

    """
    tval1 = torch.as_tensor(val1, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    tval2 = torch.as_tensor(val2, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    
    return torch.arcsin(torch.sin((tval1 - tval2) * np.pi / period)) * period / np.pi


def cyclic_difference_np(val1, val2, period):
    """Applies the difference operation between two values with cyclic
    boundary conditions.

    """
    tval1 = torch.as_tensor(val1, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    tval2 = torch.as_tensor(val2, device=AP_config.ap_device, dtype=AP_config.ap_dtype)
    return np.arcsin(np.sin((tval1 - tval2) * np.pi / period)) * period / np.pi



from typing import List, Optional

import torch
import numpy as np
from torch.nn.functional import avg_pool2d

from .image_object import Image, Image_List
from .jacobian_image import Jacobian_Image, Jacobian_Image_List
from .model_image import Model_Image, Model_Image_List
from .psf_image import PSF_Image
from astropy.io import fits
from .. import AP_config
from ..errors import SpecificationConflict, InvalidImage

__all__ = ["Target_Image", "Target_Image_List"]


class Target_Image(Image):
    """Image object which represents the data to be fit by a model. It can
    include a variance image, mask, and PSF as anciliary data which
    describes the target image.

    Target images are a basic unit of data in `AstroPhot`, they store
    the information collected from telescopes for which models are to
    be fit. There is minimial functionality in the `Target_Image`
    object itself, it is mostly defined in terms of how other objects
    interact with it.

    Basic usage:

    .. code-block:: python

      import astrophot as ap

      # Create target image
      image = ap.image.Target_Image(
          data = <pixel data>,
          wcs = <astropy WCS object>,
          variance = <pixel uncertainties>,
          psf = <point spread function as PSF_Image object>,
          mask = <pixels to ignore>,
      )

      # Display the data
      fig, ax = plt.subplots()
      ap.plots.target_image(fig, ax, image)
      plt.show()

      # Save the image
      image.save("mytarget.fits")

      # Load the image
      image2 = ap.image.Target_Image(filename = "mytarget.fits")

      # Make low resolution version
      lowrez = image.reduce(2)

    Some important information to keep in mind. First, providing an
    `astropy WCS` object is the best way to keep track of coordinates
    and pixel scale properties, especially when dealing with
    multi-band data. If images have relative positioning, rotation,
    pixel sizes, field of view this will all be handled automatically
    by taking advantage of `WCS` objects. Second, Providing accurate
    variance (or weight) maps is critical to getting a good fit to the
    data. This is a very common source of issues so it is worthwhile
    to review literature on how best to construct such a map. A good
    starting place is the FAQ for GALFIT:
    https://users.obs.carnegiescience.edu/peng/work/galfit/CHI2.html
    which is an excellent resource for all things image modeling. Just
    note that `AstroPhot` uses variance or weight maps, not sigma
    images. `AstroPhot` will not crete a variance map for the user, by
    default it will just assume uniform variance which is rarely
    accurate. Third, The PSF pixelscale must be a multiple of the
    image pixelscale. So if the image has a pixelscale of 1 then the
    PSF must have a pixelscale of 1, 1/2, 1/3, etc for anything to
    work out. Note that if the PSF pixelscale is finer than the image,
    then all modelling will be done at the higher resolution. This is
    recommended for accuracy though it can mean higher memory
    consumption.

    """

    image_count = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.has_weight and "weight" in kwargs:
            self.set_weight(kwargs.get("weight", None))
        elif not self.has_variance and "variance" in kwargs:
            self.set_variance(kwargs.get("variance", None))
        if not self.has_mask:
            self.set_mask(kwargs.get("mask", None))
        if not self.has_psf:
            self.set_psf(kwargs.get("psf", None), kwargs.get("psf_upscale", 1))

        # Set nan pixels to be masked automatically
        if torch.any(torch.isnan(self.data)).item():
            self.set_mask(torch.logical_or(self.mask, torch.isnan(self.data)))

    @property
    def standard_deviation(self):
        """Stores the standard deviation of the image pixels. This represents
        the uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the standard
        deviation is not known, a tensor of ones will be created to
        stand in as the standard deviation values.

        The standard deviation is not stored directly, instead it is
        computed as :math:`\\sqrt{1/W}` where :math:`W` is the
        weights.

        """
        if self.has_variance:
            return torch.sqrt(self.variance)
        return torch.ones_like(self.data)

    @property
    def variance(self):
        """Stores the variance of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the variance
        is not known, a tensor of ones will be created to stand in as
        the variance values.

        The variance is not stored directly, instead it is
        computed as :math:`\\frac{1}{W}` where :math:`W` is the
        weights.

        """
        if self.has_variance:
            return torch.where(self._weight == 0, torch.inf, 1 / self._weight)
        return torch.ones_like(self.data)

    @variance.setter
    def variance(self, variance):
        self.set_variance(variance)

    @property
    def has_variance(self):
        """Returns True when the image object has stored variance values. If
        this is False and the variance property is called then a
        tensor of ones will be returned.

        """
        try:
            return self._weight is not None
        except AttributeError:
            return False

    @property
    def weight(self):
        """Stores the weight of the image pixels. This represents the
        uncertainty in each pixel value. It should always have the
        same shape as the image data. In the case where the weight
        is not known, a tensor of ones will be created to stand in as
        the weight values.

        The weights are used to proprtionately scale residuals in the
        likelihood. Most commonly this shows up as a :math:`\\chi^2`
        like:

        .. math::

          \\chi^2 = (\\vec{y} - \\vec{f(\\theta)})^TW(\\vec{y} - \\vec{f(\\theta)})

        which can be optimized to find parameter values. Using the
        Jacobian, which in this case is the derivative of every pixel
        wrt every parameter, the weight matrix also appears in the
        gradient:

        .. math::

          \\vec{g} = J^TW(\\vec{y} - \\vec{f(\\theta)})

        and the hessian approximation used in Levenberg-Marquardt:

        .. math::

          H \\approx J^TWJ

        """
        if self.has_weight:
            return self._weight
        return torch.ones_like(self.data)

    @weight.setter
    def weight(self, weight):
        self.set_weight(weight)

    @property
    def has_weight(self):
        """Returns True when the image object has stored weight values. If
        this is False and the weight property is called then a
        tensor of ones will be returned.

        """
        try:
            return self._weight is not None
        except AttributeError:
            self._weight = None
            return False

    @property
    def mask(self):
        """The mask stores a tensor of boolean values which indicate any
        pixels to be ignored. These pixels will be skipped in
        likelihood evaluations and in parameter optimization. It is
        common practice to mask pixels with pathological values such
        as due to cosmic rays or satelites passing through the image.

        In a mask, a True value indicates that the pixel is masked and
        should be ignored. False indicates a normal pixel which will
        inter into most calculaitons.

        If no mask is provided, all pixels are assumed valid.

        """
        if self.has_mask:
            return self._mask
        return torch.zeros_like(self.data, dtype=torch.bool)

    @mask.setter
    def mask(self, mask):
        self.set_mask(mask)

    @property
    def has_mask(self):
        """
        Single boolean to indicate if a mask has been provided by the user.
        """
        try:
            return self._mask is not None
        except AttributeError:
            return False

    @property
    def psf(self):
        """Stores the point-spread-function for this target. This should be a
        `PSF_Image` object which represents the scattering of a point
        source of light. It can also be an `AstroPhot_Model` object
        which will contribute its own parameters to an optimization
        problem.

        The PSF stored for a `Target_Image` object is passed to all
        models applied to that target which have a `psf_mode` that is
        not `none`. This means they will all use the same PSF
        model. If one wishes to define a variable PSF across an image,
        then they should pass the PSF objects to the `AstroPhot_Model`'s
        directly instead of to a `Target_Image`.

        Raises:

          AttributeError: if this is called without a PSF defined

        """
        if self.has_psf:
            return self._psf
        raise AttributeError("This image does not have a PSF")

    @psf.setter
    def psf(self, psf):
        self.set_psf(psf)

    @property
    def has_psf(self):
        try:
            return self._psf is not None
        except AttributeError:
            return False

    def set_variance(self, variance):
        """
        Provide a variance tensor for the image. Variance is equal to $\\sigma^2$. This should have the same shape as the data.
        """
        if variance is None:
            self._weight = None
            return
        self.set_weight(1 / variance)

    def set_weight(self, weight):
        """Provide a weight tensor for the image. Weight is equal to $\\frac{1}{\\sigma^2}$. This should have the same
        shape as the data.

        """
        if weight is None:
            self._weight = None
            return
        if weight.shape != self.data.shape:
            raise SpecificationConflict(
                f"weight/variance must have same shape as data ({weight.shape} vs {self.data.shape})"
            )
        self._weight = (
            weight.to(dtype=AP_config.ap_dtype, device=AP_config.ap_device)
            if isinstance(weight, torch.Tensor)
            else torch.as_tensor(weight, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        )

    def set_psf(self, psf, psf_upscale=1):
        """Provide a psf for the `Target_Image`. This is stored and passed to
        models which need to be convolved.

        The PSF doesn't need to have the same pixelscale as the
        image. It should be some multiple of the resolution of the
        `Target_Image` though. So if the image has a pixelscale of 1,
        the psf may have a pixelscale of 1, 1/2, 1/3, 1/4 and so on.

        """
        if psf is None:
            self._psf = None
            return
        if isinstance(psf, PSF_Image):
            self._psf = psf
            return

        self._psf = PSF_Image(
            data=psf,
            psf_upscale=psf_upscale,
            pixelscale=self.pixelscale / psf_upscale,
            identity=self.identity,
        )

    def set_mask(self, mask):
        """
        Set the boolean mask which will indicate which pixels to ignore. A mask value of True means the pixel will be ignored.
        """
        if mask is None:
            self._mask = None
            return
        if mask.shape != self.data.shape:
            raise SpecificationConflict(
                f"mask must have same shape as data ({mask.shape} vs {self.data.shape})"
            )
        self._mask = (
            mask.to(dtype=torch.bool, device=AP_config.ap_device)
            if isinstance(mask, torch.Tensor)
            else torch.as_tensor(mask, dtype=torch.bool, device=AP_config.ap_device)
        )

    def to(self, dtype=None, device=None):
        """Converts the stored `Target_Image` data, variance, psf, etc to a
        given data type and device.

        """
        super().to(dtype=dtype, device=device)
        if dtype is not None:
            dtype = AP_config.ap_dtype
        if device is not None:
            device = AP_config.ap_device

        if self.has_weight:
            self._weight = self._weight.to(dtype=dtype, device=device)
        if self.has_psf:
            self._psf = self._psf.to(dtype=dtype, device=device)
        if self.has_mask:
            self._mask = self.mask.to(dtype=torch.bool, device=device)
        return self

    def or_mask(self, mask):
        """
        Combines the currently stored mask with a provided new mask using the boolean `or` operator.
        """
        self._mask = torch.logical_or(self.mask, mask)

    def and_mask(self, mask):
        """
        Combines the currently stored mask with a provided new mask using the boolean `and` operator.
        """
        self._mask = torch.logical_and(self.mask, mask)

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        return super().copy(
            mask=self._mask,
            psf=self._psf,
            weight=self._weight,
            **kwargs,
        )

    def blank_copy(self, **kwargs):
        """Produces a blank copy of the image which has the same properties
        except that its data is not filled with zeros.

        """
        return super().blank_copy(mask=self._mask, psf=self._psf, **kwargs)

    def get_window(self, window, **kwargs):
        """Get a sub-region of the image as defined by a window on the sky."""
        indices = self.window.get_self_indices(window)
        return super().get_window(
            window=window,
            weight=self._weight[indices] if self.has_weight else None,
            mask=self._mask[indices] if self.has_mask else None,
            psf=self._psf,
            **kwargs,
        )

    def jacobian_image(
        self,
        parameters: Optional[List[str]] = None,
        data: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """
        Construct a blank `Jacobian_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        if parameters is None:
            data = None
            parameters = []
        elif data is None:
            data = torch.zeros(
                (*self.data.shape, len(parameters)),
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
        return Jacobian_Image(
            parameters=parameters,
            target_identity=self.identity,
            data=data,
            header=self.header,
            **kwargs,
        )

    def model_image(self, data: Optional[torch.Tensor] = None, **kwargs):
        """
        Construct a blank `Model_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        return Model_Image(
            data=torch.zeros_like(self.data) if data is None else data,
            header=self.header,
            target_identity=self.identity,
            **kwargs,
        )

    def reduce(self, scale, **kwargs):
        """Returns a new `Target_Image` object with a reduced resolution
        compared to the current image. `scale` should be an integer
        indicating how much to reduce the resolution. If the
        `Target_Image` was originally (48,48) pixels across with a
        pixelscale of 1 and `reduce(2)` is called then the image will
        be (24,24) pixels and the pixelscale will be 2. If `reduce(3)`
        is called then the returned image will be (16,16) pixels
        across and the pixelscale will be 3.

        """
        MS = self.data.shape[0] // scale
        NS = self.data.shape[1] // scale

        return super().reduce(
            scale=scale,
            variance=(
                self.variance[: MS * scale, : NS * scale]
                .reshape(MS, scale, NS, scale)
                .sum(axis=(1, 3))
                if self.has_variance
                else None
            ),
            mask=(
                self.mask[: MS * scale, : NS * scale]
                .reshape(MS, scale, NS, scale)
                .amax(axis=(1, 3))
                if self.has_mask
                else None
            ),
            psf=self.psf.reduce(scale) if self.has_psf else None,
            **kwargs,
        )

    def expand(self, padding):
        """
        `Target_Image` doesn't have expand yet.
        """
        raise NotImplementedError("expand not available for Target_Image yet")

    def get_state(self):
        state = super().get_state()

        if self.has_weight:
            state["weight"] = self.weight.detach().cpu().tolist()
        if self.has_mask:
            state["mask"] = self.mask.detach().cpu().tolist()
        if self.has_psf:
            state["psf"] = self.psf.get_state()

        return state

    def set_state(self, state):
        super().set_state(state)

        self.weight = state.get("weight", None)
        self.mask = state.get("mask", None)
        if "psf" in state:
            self.psf = PSF_Image(state=state["psf"])

    def get_fits_state(self):
        states = super().get_fits_state()
        if self.has_weight:
            states.append(
                {
                    "DATA": self.weight.detach().cpu().numpy(),
                    "HEADER": {"IMAGE": "WEIGHT"},
                }
            )
        if self.has_mask:
            states.append(
                {
                    "DATA": self.mask.detach().cpu().numpy().astype(int),
                    "HEADER": {"IMAGE": "MASK"},
                }
            )
        if self.has_psf:
            states += self.psf.get_fits_state()

        return states

    def set_fits_state(self, states):
        super().set_fits_state(states)
        for state in states:
            if state["HEADER"]["IMAGE"] == "WEIGHT":
                self.weight = np.array(state["DATA"], dtype=np.float64)
            if state["HEADER"]["IMAGE"] == "mask":
                self.mask = np.array(state["DATA"], dtype=bool)
            if state["HEADER"]["IMAGE"] == "PSF":
                self.psf = PSF_Image(fits_state=states)


class Target_Image_List(Image_List, Target_Image):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not all(isinstance(image, Target_Image) for image in self.image_list):
            raise InvalidImage(
                f"Target_Image_List can only hold Target_Image objects, not {tuple(type(image) for image in self.image_list)}"
            )

    @property
    def variance(self):
        return tuple(image.variance for image in self.image_list)

    @variance.setter
    def variance(self, variance):
        for image, var in zip(self.image_list, variance):
            image.set_variance(var)

    @property
    def has_variance(self):
        return any(image.has_variance for image in self.image_list)

    @property
    def weight(self):
        return tuple(image.weight for image in self.image_list)

    @weight.setter
    def weight(self, weight):
        for image, wgt in zip(self.image_list, weight):
            image.set_weight(wgt)

    @property
    def has_weight(self):
        return any(image.has_weight for image in self.image_list)

    def jacobian_image(self, parameters: List[str], data: Optional[List[torch.Tensor]] = None):
        if data is None:
            data = [None] * len(self.image_list)
        return Jacobian_Image_List(
            list(image.jacobian_image(parameters, dat) for image, dat in zip(self.image_list, data))
        )

    def model_image(self, data: Optional[List[torch.Tensor]] = None):
        if data is None:
            data = [None] * len(self.image_list)
        return Model_Image_List(
            list(image.model_image(data=dat) for image, dat in zip(self.image_list, data))
        )

    def match_indices(self, other):
        indices = []
        if isinstance(other, Target_Image_List):
            for other_image in other.image_list:
                for isi, self_image in enumerate(self.image_list):
                    if other_image.identity == self_image.identity:
                        indices.append(isi)
                        break
                else:
                    indices.append(None)
        elif isinstance(other, Target_Image):
            for isi, self_image in enumerate(self.image_list):
                if other.identity == self_image.identity:
                    indices = isi
                    break
            else:
                indices = None
        return indices

    def __isub__(self, other):
        if isinstance(other, Target_Image_List):
            for other_image in other.image_list:
                for self_image in self.image_list:
                    if other_image.identity == self_image.identity:
                        self_image -= other_image
                        break
                else:
                    self.image_list.append(other_image)
        elif isinstance(other, Target_Image):
            for self_image in self.image_list:
                if other.identity == self_image.identity:
                    self_image -= other
                    break
        elif isinstance(other, Model_Image_List):
            for other_image in other.image_list:
                for self_image in self.image_list:
                    if other_image.target_identity == self_image.identity:
                        self_image -= other_image
                        break
        elif isinstance(other, Model_Image):
            for self_image in self.image_list:
                if other.target_identity == self_image.identity:
                    self_image -= other
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image -= other_image
        return self

    def __iadd__(self, other):
        if isinstance(other, Target_Image_List):
            for other_image in other.image_list:
                for self_image in self.image_list:
                    if other_image.identity == self_image.identity:
                        self_image += other_image
                        break
                else:
                    self.image_list.append(other_image)
        elif isinstance(other, Target_Image):
            for self_image in self.image_list:
                if other.identity == self_image.identity:
                    self_image += other
        elif isinstance(other, Model_Image_List):
            for other_image in other.image_list:
                for self_image in self.image_list:
                    if other_image.target_identity == self_image.identity:
                        self_image += other_image
                        break
        elif isinstance(other, Model_Image):
            for self_image in self.image_list:
                if other.target_identity == self_image.identity:
                    self_image += other
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image += other_image
        return self

    @property
    def mask(self):
        return tuple(image.mask for image in self.image_list)

    @mask.setter
    def mask(self, mask):
        for image, M in zip(self.image_list, mask):
            image.set_mask(M)

    @property
    def has_mask(self):
        return any(image.has_mask for image in self.image_list)

    @property
    def psf(self):
        return tuple(image.psf for image in self.image_list)

    @psf.setter
    def psf(self, psf):
        for image, P in zip(self.image_list, psf):
            image.set_psf(P)

    @property
    def has_psf(self):
        return any(image.has_psf for image in self.image_list)

    @property
    def psf_border(self):
        return tuple(image.psf_border for image in self.image_list)

    @property
    def psf_border_int(self):
        return tuple(image.psf_border_int for image in self.image_list)

    def set_variance(self, variance, img):
        self.image_list[img].set_variance(variance)

    def set_psf(self, psf, img):
        self.image_list[img].set_psf(psf)

    def set_mask(self, mask, img):
        self.image_list[img].set_mask(mask)

    def or_mask(self, mask):
        raise NotImplementedError()

    def and_mask(self, mask):
        raise NotImplementedError()



from .model_object import Component_Model

__all__ = ["Sky_Model"]


class Sky_Model(Component_Model):
    """prototype class for any sky backgorund model. This simply imposes
    that the center is a locked parameter, not involved in the
    fit. Also, a sky model object has no psf mode or integration mode
    by default since it is intended to change over much larger spatial
    scales than the psf or pixel size.

    """

    model_type = f"sky {Component_Model.model_type}"
    parameter_specs = {
        "center": {"units": "arcsec", "locked": True, "uncertainty": 0.0},
    }
    useable = False

    @property
    def psf_mode(self):
        return "none"

    @psf_mode.setter
    def psf_mode(self, val):
        pass

    @property
    def integrate_mode(self):
        return "none"

    @integrate_mode.setter
    def integrate_mode(self, val):
        pass



import numpy as np
import torch

from .galaxy_model_object import Galaxy_Model
from ..param import Parameter_Node
from ..utils.interpolate import cubic_spline_torch
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.conversions.coordinates import Axis_Ratio_Cartesian

__all__ = ["Ray_Galaxy"]


class Ray_Galaxy(Galaxy_Model):
    """Variant of a galaxy model which defines multiple radial models
    seprarately along some number of rays projected from the galaxy
    center. These rays smoothly transition from one to another along
    angles theta. The ray transition uses a cosine smoothing function
    which depends on the number of rays, for example with two rays the
    brightness would be:

    I(R,theta) = I1(R)*cos(theta % pi) + I2(R)*cos((theta + pi/2) % pi)

    Where I(R,theta) is the brightness function in polar coordinates,
    R is the semi-major axis, theta is the polar angle (defined after
    galaxy axis ratio is applied), I1(R) is the first brightness
    profile, % is the modulo operator, and I2 is the second brightness
    profile. The ray model defines no extra parameters, though now
    every model parameter related to the brightness profile gains an
    extra dimension for the ray number. Also a new input can be given
    when instantiating the ray model: "rays" which is an integer for
    the number of rays.

    """

    model_type = f"ray {Galaxy_Model.model_type}"
    special_kwargs = Galaxy_Model.special_kwargs + ["rays"]
    rays = 2
    track_attrs = Galaxy_Model.track_attrs + ["rays"]
    useable = False

    def __init__(self, *args, **kwargs):
        self.symmetric_rays = True
        super().__init__(*args, **kwargs)
        self.rays = kwargs.get("rays", Ray_Galaxy.rays)

    @default_internal
    def polar_model(self, R, T, image=None, parameters=None):
        model = torch.zeros_like(R)
        if self.rays % 2 == 0 and self.symmetric_rays:
            for r in range(self.rays):
                angles = (T - (r * np.pi / self.rays)) % np.pi
                indices = torch.logical_or(
                    angles < (np.pi / self.rays),
                    angles >= (np.pi * (1 - 1 / self.rays)),
                )
                weight = (torch.cos(angles[indices] * self.rays) + 1) / 2
                model[indices] += weight * self.iradial_model(r, R[indices], image)
        elif self.rays % 2 == 1 and self.symmetric_rays:
            for r in range(self.rays):
                angles = (T - (r * np.pi / self.rays)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (np.pi / self.rays),
                    angles >= (np.pi * (2 - 1 / self.rays)),
                )
                weight = (torch.cos(angles[indices] * self.rays) + 1) / 2
                model[indices] += weight * self.iradial_model(r, R[indices], image)
                angles = (T - (np.pi + r * np.pi / self.rays)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (np.pi / self.rays),
                    angles >= (np.pi * (2 - 1 / self.rays)),
                )
                weight = (torch.cos(angles[indices] * self.rays) + 1) / 2
                model[indices] += weight * self.iradial_model(r, R[indices], image)
        elif self.rays % 2 == 0 and not self.symmetric_rays:
            for r in range(self.rays):
                angles = (T - (r * 2 * np.pi / self.rays)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (2 * np.pi / self.rays),
                    angles >= (2 * np.pi * (1 - 1 / self.rays)),
                )
                weight = (torch.cos(angles[indices] * self.rays) + 1) / 2
                model[indices] += weight * self.iradial_model(r, R[indices], image)
        else:
            for r in range(self.rays):
                angles = (T - (r * 2 * np.pi / self.rays)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (2 * np.pi / self.rays),
                    angles >= (np.pi * (2 - 1 / self.rays)),
                )
                weight = (torch.cos(angles[indices] * self.rays) + 1) / 2
                model[indices] += weight * self.iradial_model(r, R[indices], image)
        return model

    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        XX, YY = self.transform_coordinates(X, Y, image, parameters)

        return self.polar_model(
            self.radius_metric(XX, YY, image=image, parameters=parameters),
            self.angular_metric(XX, YY, image=image, parameters=parameters),
            image=image,
            parameters=parameters,
        )


# class SingleRay_Galaxy(Galaxy_Model):



import numpy as np
import logging
from scipy.stats import iqr

from .ellipse import parametric_SuperEllipse, Rscale_SuperEllipse
from ..conversions.coordinates import Rotate_Cartesian_np
from ..interpolate import interpolate_Lanczos


def Sigma_Clip_Upper(v, iterations=10, nsigma=5):
    """
    Perform sigma clipping on the "v" array. Each iteration involves
    computing the median and 16-84 range, these are used to clip beyond
    "nsigma" number of sigma above the median. This is repeated for
    "iterations" number of iterations, or until convergence if None.
    """

    v2 = np.sort(v)
    i = 0
    old_lim = 0
    lim = np.inf
    while i < iterations and old_lim != lim:
        med = np.median(v2[v2 < lim])
        rng = iqr(v2[v2 < lim], rng=[16, 84]) / 2
        old_lim = lim
        lim = med + rng * nsigma
        i += 1
    return lim


def _iso_between(
    IMG,
    sma_low,
    sma_high,
    PARAMS,
    c,
    more=False,
    mask=None,
    sigmaclip=False,
    sclip_iterations=10,
    sclip_nsigma=5,
):
    if not "m" in PARAMS:
        PARAMS["m"] = None
    if not "C" in PARAMS:
        PARAMS["C"] = None
    Rlim = sma_high * (
        1.0
        if PARAMS["m"] is None
        else np.exp(sum(np.abs(PARAMS["Am"][m]) for m in range(len(PARAMS["m"]))))
    )
    ranges = [
        [max(0, int(c["x"] - Rlim - 2)), min(IMG.shape[1], int(c["x"] + Rlim + 2))],
        [max(0, int(c["y"] - Rlim - 2)), min(IMG.shape[0], int(c["y"] + Rlim + 2))],
    ]
    XX, YY = np.meshgrid(
        np.arange(ranges[0][1] - ranges[0][0], dtype=float)
        - c["x"]
        + float(ranges[0][0]),
        np.arange(ranges[1][1] - ranges[1][0], dtype=float)
        - c["y"]
        + float(ranges[1][0]),
    )
    theta = np.arctan(YY / XX) + np.pi * (XX < 0)
    RR = np.sqrt(XX ** 2 + YY ** 2)
    Fmode_Rscale = (
        1.0
        if PARAMS["m"] is None
        else Rscale_Fmodes(
            theta - PARAMS["pa"], PARAMS["m"], PARAMS["Am"], PARAMS["Phim"]
        )
    )
    SuperEllipse_Rscale = Rscale_SuperEllipse(
        theta - PARAMS["pa"], PARAMS["ellip"], 2 if PARAMS["C"] is None else PARAMS["C"]
    )
    RR /= SuperEllipse_Rscale * Fmode_Rscale
    rselect = np.logical_and(RR < sma_high, RR > sma_low)
    fluxes = IMG[ranges[1][0] : ranges[1][1], ranges[0][0] : ranges[0][1]][rselect]
    CHOOSE = None
    if not mask is None and sma_high > 5:
        CHOOSE = np.logical_not(
            mask[ranges[1][0] : ranges[1][1], ranges[0][0] : ranges[0][1]][rselect]
        )
    # Perform sigma clipping if requested
    if sigmaclip:
        sclim = Sigma_Clip_Upper(fluxes, sclip_iterations, sclip_nsigma)
        if CHOOSE is None:
            CHOOSE = fluxes < sclim
        else:
            CHOOSE = np.logical_or(CHOOSE, fluxes < sclim)
    if CHOOSE is not None and np.sum(CHOOSE) < 5:
        logging.warning(
            "Entire Isophote is Masked! R_l: %.3f, R_h: %.3f, PA: %.3f, ellip: %.3f"
            % (sma_low, sma_high, PARAMS["pa"] * 180 / np.pi, PARAMS["ellip"])
        )
        CHOOSE = np.ones(CHOOSE.shape).astype(bool)
    if CHOOSE is not None:
        countmasked = np.sum(np.logical_not(CHOOSE))
    else:
        countmasked = 0
    if more:
        if CHOOSE is not None and sma_high > 5:
            return fluxes[CHOOSE], theta[rselect][CHOOSE], countmasked
        else:
            return fluxes, theta[rselect], countmasked
    else:
        if CHOOSE is not None and sma_high > 5:
            return fluxes[CHOOSE]
        else:
            return fluxes


def _iso_extract(
    IMG,
    sma,
    PARAMS,
    c,
    more=False,
    minN=None,
    mask=None,
    interp_mask=False,
    rad_interp=30,
    interp_method="lanczos",
    interp_window=5,
    sigmaclip=False,
    sclip_iterations=10,
    sclip_nsigma=5,
):
    """
    Internal, basic function for extracting the pixel fluxes along an isophote
    """
    if not "m" in PARAMS:
        PARAMS["m"] = None
    if not "C" in PARAMS:
        PARAMS["C"] = None
    N = max(15, int(0.9 * 2 * np.pi * sma))
    if not minN is None:
        N = max(minN, N)
    # points along ellipse to evaluate
    theta = np.linspace(0, 2 * np.pi * (1.0 - 1.0 / N), N)
    theta = np.arctan(PARAMS["q"] * np.tan(theta)) + np.pi * (np.cos(theta) < 0)
    Fmode_Rscale = (
        1.0
        if PARAMS["m"] is None
        else Rscale_Fmodes(theta, PARAMS["m"], PARAMS["Am"], PARAMS["Phim"])
    )
    R = sma * Fmode_Rscale
    # Define ellipse
    X, Y = parametric_SuperEllipse(
        theta, 1.0 - PARAMS["q"], 2 if PARAMS["C"] is None else PARAMS["C"]
    )
    X, Y = R * X, R * Y
    # rotate ellipse by PA
    X, Y = Rotate_Cartesian_np(PARAMS["pa"], X, Y)
    theta = (theta + PARAMS["pa"]) % (2 * np.pi)
    # shift center
    X, Y = X + c["x"], Y + c["y"]

    # Reject samples from outside the image
    BORDER = np.logical_and(
        np.logical_and(X >= 0, X < (IMG.shape[1] - 1)),
        np.logical_and(Y >= 0, Y < (IMG.shape[0] - 1)),
    )
    if not np.all(BORDER):
        X = X[BORDER]
        Y = Y[BORDER]
        theta = theta[BORDER]

    Rlim = np.max(R)
    if Rlim < rad_interp:
        box = [
            [max(0, int(c["x"] - Rlim - 5)), min(IMG.shape[1], int(c["x"] + Rlim + 5))],
            [max(0, int(c["y"] - Rlim - 5)), min(IMG.shape[0], int(c["y"] + Rlim + 5))],
        ]
        if interp_method == "bicubic":
            flux = interpolate_bicubic(
                IMG[box[1][0] : box[1][1], box[0][0] : box[0][1]],
                X - box[0][0],
                Y - box[1][0],
            )
        elif interp_method == "lanczos":
            flux = interpolate_Lanczos(IMG, X, Y, interp_window)
        else:
            raise ValueError(
                "Unknown interpolate method %s. Should be one of lanczos or bicubic"
                % interp_method
            )
    else:
        # round to integers and sample pixels values
        flux = IMG[np.rint(Y).astype(np.int32), np.rint(X).astype(np.int32)]
    # CHOOSE holds bolean array for which flux values to keep, initialized as None for no clipping
    CHOOSE = None
    # Mask pixels if a mask is given
    if not mask is None:
        CHOOSE = np.logical_not(
            mask[np.rint(Y).astype(np.int32), np.rint(X).astype(np.int32)]
        )
    # Perform sigma clipping if requested
    if sigmaclip and len(flux) > 30:
        sclim = Sigma_Clip_Upper(flux, sclip_iterations, sclip_nsigma)
        if CHOOSE is None:
            CHOOSE = flux < sclim
        else:
            CHOOSE = np.logical_or(CHOOSE, flux < sclim)
    # Dont clip pixels if that removes all of the pixels
    countmasked = 0
    if not CHOOSE is None and np.sum(CHOOSE) <= 0:
        logging.warning(
            "Entire Isophote was Masked! R: %.3f, PA: %.3f, q: %.3f"
            % (sma, PARAMS["pa"] * 180 / np.pi, PARAMS["q"])
        )
        # Interpolate clipped flux values if requested
    elif not CHOOSE is None and interp_mask:
        flux[np.logical_not(CHOOSE)] = np.interp(
            theta[np.logical_not(CHOOSE)], theta[CHOOSE], flux[CHOOSE], period=2 * np.pi
        )
        # simply remove all clipped pixels if user doesn't reqest another option
    elif not CHOOSE is None:
        flux = flux[CHOOSE]
        theta = theta[CHOOSE]
        countmasked = np.sum(np.logical_not(CHOOSE))

    # Return just the flux values, or flux and angle values
    if more:
        return flux, theta, countmasked
    else:
        return flux


def _iso_line(IMG, length, width, pa, c, more=False):
    start = np.array([c["x"], c["y"]])
    end = start + length * np.array([np.cos(pa), np.sin(pa)])

    ranges = [
        [
            max(0, int(min(start[0], end[0]) - 2)),
            min(IMG.shape[1], int(max(start[0], end[0]) + 2)),
        ],
        [
            max(0, int(min(start[1], end[1]) - 2)),
            min(IMG.shape[0], int(max(start[1], end[1]) + 2)),
        ],
    ]
    XX, YY = np.meshgrid(
        np.arange(ranges[0][1] - ranges[0][0], dtype=float),
        np.arange(ranges[1][1] - ranges[1][0], dtype=float),
    )
    XX -= c["x"] - float(ranges[0][0])
    YY -= c["y"] - float(ranges[1][0])
    XX, YY = (XX * np.cos(-pa) - YY * np.sin(-pa), XX * np.sin(-pa) + YY * np.cos(-pa))

    lselect = np.logical_and.reduce(
        (XX >= -0.5, XX <= length, np.abs(YY) <= (width / 2))
    )
    flux = IMG[ranges[1][0] : ranges[1][1], ranges[0][0] : ranges[0][1]][lselect]

    if more:
        return flux, XX[lselect], YY[lselect]
    else:
        return flux, XX[lselect]



# Levenberg-Marquardt algorithm
import os
from time import time
from typing import List, Callable, Optional, Union, Sequence, Any

import torch
from torch.autograd.functional import jacobian
import numpy as np
import matplotlib.pyplot as plt

from .base import BaseOptimizer
from .. import AP_config

__all__ = ["oldLM", "LM_Constraint"]


@torch.no_grad()
@torch.jit.script
def Broyden_step(J, h, Yp, Yph):
    delta = torch.matmul(J, h)
    # avoid constructing a second giant jacobian matrix, instead go one row at a time
    for j in range(J.shape[1]):
        J[:, j] += (Yph - Yp - delta) * h[j] / torch.linalg.norm(h)
    return J


class oldLM(BaseOptimizer):
    """based heavily on:
    @article{gavin2019levenberg,
        title={The Levenberg-Marquardt algorithm for nonlinear least squares curve-fitting problems},
        author={Gavin, Henri P},
        journal={Department of Civil and Environmental Engineering, Duke University},
        volume={19},
        year={2019}
    }

    The Levenberg-Marquardt algorithm bridges the gap between a
    gradient descent optimizer and a Newton's Method optimizer. The
    Hessian for the Newton's Method update is too complex to evaluate
    with automatic differentiation (memory scales roughly as
    parameters^2 * pixels^2) and so an approximation is made using the
    Jacobian of the image pixels wrt to the parameters of the
    model. Automatic differentiation provides an exact Jacobian as
    opposed to a finite differences approximation.

    Once a Hessian H and gradient G have been determined, the update
    step is defined as h which is the solution to the linear equation:

    (H + L*I)h = G

    where L is the Levenberg-Marquardt damping parameter and I is the
    identity matrix. For small L this is just the Newton's method, for
    large L this is just a small gradient descent step (approximately
    h = grad/L). The method implimented is modified from Gavin 2019.

    Args:
        model (AstroPhot_Model): object with which to perform optimization
        initial_state (Optional[Sequence]): an initial state for optimization
        epsilon4 (Optional[float]): approximation accuracy requirement, for any rho < epsilon4 the step will be rejected. Default 0.1
        epsilon5 (Optional[float]): numerical stability factor, added to the diagonal of the Hessian. Default 1e-8
        constraints (Optional[Union[LM_Constraint,tuple[LM_Constraint]]]): Constraint objects which control the fitting process.
        L0 (Optional[float]): initial value for L factor in (H +L*I)h = G. Default 1.
        Lup (Optional[float]): amount to increase L when rejecting an update step. Default 11.
        Ldn (Optional[float]): amount to decrease L when accetping an update step. Default 9.

    """

    def __init__(
        self,
        model: "AstroPhot_Model",
        initial_state: Sequence = None,
        max_iter: int = 100,
        fit_parameters_identity: Optional[tuple] = None,
        **kwargs,
    ):
        super().__init__(
            model,
            initial_state,
            max_iter=max_iter,
            fit_parameters_identity=fit_parameters_identity,
            **kwargs,
        )

        # Set optimizer parameters
        self.epsilon4 = kwargs.get("epsilon4", 0.1)
        self.epsilon5 = kwargs.get("epsilon5", 1e-8)
        self.Lup = kwargs.get("Lup", 11.0)
        self.Ldn = kwargs.get("Ldn", 9.0)
        self.L = kwargs.get("L0", 1e-3)
        self.use_broyden = kwargs.get("use_broyden", False)

        # Initialize optimizer atributes
        self.Y = self.model.target[self.fit_window].flatten("data")
        #        1 / sigma^2
        self.W = (
            1.0 / self.model.target[self.fit_window].flatten("variance")
            if model.target.has_variance
            else 1.0
        )
        #          # pixels      # parameters
        self.ndf = len(self.Y) - len(self.current_state)
        self.J = None
        self.full_jac = False
        self.current_Y = None
        self.prev_Y = [None, None]
        if self.model.target.has_mask:
            self.mask = self.model.target[self.fit_window].flatten("mask")
            # subtract masked pixels from degrees of freedom
            self.ndf -= torch.sum(self.mask)
        self.L_history = []
        self.decision_history = []
        self.rho_history = []
        self._count_converged = 0
        self.ndf = kwargs.get("ndf", self.ndf)
        self._covariance_matrix = None

        # update attributes with constraints
        self.constraints = kwargs.get("constraints", None)
        if self.constraints is not None and isinstance(self.constraints, LM_Constraint):
            self.constraints = (self.constraints,)

        if self.constraints is not None:
            for con in self.constraints:
                self.Y = torch.cat((self.Y, con.reference_value))
                self.W = torch.cat((self.W, 1 / con.weight))
                self.ndf -= con.reduce_ndf
                if self.model.target.has_mask:
                    self.mask = torch.cat(
                        (
                            self.mask,
                            torch.zeros_like(con.reference_value, dtype=torch.bool),
                        )
                    )

    def L_up(self, Lup=None):
        if Lup is None:
            Lup = self.Lup
        self.L = min(1e9, self.L * Lup)

    def L_dn(self, Ldn=None):
        if Ldn is None:
            Ldn = self.Ldn
        self.L = max(1e-9, self.L / Ldn)

    def step(self, current_state=None) -> None:
        """
        Levenberg-Marquardt update step
        """
        if current_state is not None:
            self.current_state = current_state

        if self.iteration > 0:
            if self.verbose > 0:
                AP_config.ap_logger.info("---------iter---------")
        else:
            if self.verbose > 0:
                AP_config.ap_logger.info("---------init---------")

        h = self.update_h()
        if self.verbose > 1:
            AP_config.ap_logger.info(f"h: {h.detach().cpu().numpy()}")

        self.update_Yp(h)
        loss = self.update_chi2()
        if self.verbose > 0:
            AP_config.ap_logger.info(f"LM loss: {loss.item()}")

        if self.iteration == 0:
            self.prev_Y[1] = self.current_Y
        self.loss_history.append(loss.detach().cpu().item())
        self.L_history.append(self.L)
        self.lambda_history.append(
            np.copy((self.current_state + h).detach().cpu().numpy())
        )

        if self.iteration > 0 and not torch.isfinite(loss):
            if self.verbose > 0:
                AP_config.ap_logger.warning("nan loss")
            self.decision_history.append("nan")
            self.rho_history.append(None)
            self._count_reject += 1
            self.iteration += 1
            self.L_up()
            return
        elif self.iteration > 0:
            lossmin = np.nanmin(self.loss_history[:-1])
            rho = self.rho(lossmin, loss, h)
            if self.verbose > 1:
                AP_config.ap_logger.debug(
                    f"LM loss: {loss.item()}, best loss: {np.nanmin(self.loss_history[:-1])}, loss diff: {np.nanmin(self.loss_history[:-1]) - loss.item()}, L: {self.L}"
                )
            self.rho_history.append(rho)
            if self.verbose > 1:
                AP_config.ap_logger.debug(f"rho: {rho.item()}")

            if rho > self.epsilon4:
                if self.verbose > 0:
                    AP_config.ap_logger.info("accept")
                self.decision_history.append("accept")
                self.prev_Y[0] = self.prev_Y[1]
                self.prev_Y[1] = torch.clone(self.current_Y)
                self.current_state += h
                self.L_dn()
                self._count_reject = 0
                if 0 < ((lossmin - loss) / loss) < self.relative_tolerance:
                    self._count_finish += 1
                else:
                    self._count_finish = 0
            else:
                if self.verbose > 0:
                    AP_config.ap_logger.info("reject")
                self.decision_history.append("reject")
                self.L_up()
                self._count_reject += 1
                return
        else:
            self.decision_history.append("init")
            self.rho_history.append(None)

        if (
            (not self.use_broyden)
            or self.J is None
            or self.iteration < 2
            or "reset" in self.decision_history[-2:]
            or rho < self.epsilon4
            or self._count_reject > 0
            or self.iteration >= (2 * len(self.current_state))
            or self.decision_history[-1] == "nan"
        ):
            if self.verbose > 1:
                AP_config.ap_logger.debug("full jac")
            self.update_J_AD()
        else:
            if self.verbose > 1:
                AP_config.ap_logger.debug("Broyden jac")
            self.update_J_Broyden(h, self.prev_Y[0], self.current_Y)

        self.update_hess()
        self.update_grad(self.prev_Y[1])
        self.iteration += 1

    def fit(self):
        self.iteration = 0
        self._count_reject = 0
        self._count_finish = 0
        self.grad_only = False

        start_fit = time()
        try:
            while True:
                if self.verbose > 0:
                    AP_config.ap_logger.info(f"L: {self.L}")

                # take LM step
                self.step()

                # Save the state of the model
                if (
                    self.save_steps is not None
                    and self.decision_history[-1] == "accept"
                ):
                    self.model.save(
                        os.path.join(
                            self.save_steps,
                            f"{self.model.name}_Iteration_{self.iteration:03d}.yaml",
                        )
                    )

                lam, L, loss = self.progress_history()

                # Check for convergence
                if (
                    self.decision_history.count("accept") > 2
                    and self.decision_history[-1] == "accept"
                    and L[-1] < 0.1
                    and ((loss[-2] - loss[-1]) / loss[-1])
                    < (self.relative_tolerance / 10)
                ):
                    self._count_converged += 1
                elif self.iteration >= self.max_iter:
                    self.message = (
                        self.message + f"fail max iterations reached: {self.iteration}"
                    )
                    break
                elif not torch.all(torch.isfinite(self.current_state)):
                    self.message = self.message + "fail non-finite step taken"
                    break
                elif (
                    self.L >= (1e9 - 1)
                    and self._count_reject >= 8
                    and not self.take_low_rho_step()
                ):
                    self.message = (
                        self.message
                        + "fail by immobility, unable to find improvement or even small bad step"
                    )
                    break
                if self._count_converged >= 3:
                    self.message = self.message + "success"
                    break
                lam, L, loss = self.accept_history()
                if len(loss) >= 10:
                    loss10 = np.array(loss[-10:])
                    if (
                        np.all(
                            np.abs((loss10[0] - loss10[-1]) / loss10[-1])
                            < self.relative_tolerance
                        )
                        and L[-1] < 0.1
                    ):
                        self.message = self.message + "success"
                        break
                    if (
                        np.all(
                            np.abs((loss10[0] - loss10[-1]) / loss10[-1])
                            < self.relative_tolerance
                        )
                        and L[-1] >= 0.1
                    ):
                        self.message = (
                            self.message
                            + "fail by immobility, possible bad area of parameter space."
                        )
                        break
        except KeyboardInterrupt:
            self.message = self.message + "fail interrupted"

        if self.message.startswith("fail") and self._count_finish > 0:
            self.message = (
                self.message
                + ". possibly converged to numerical precision and could not make a better step."
            )
        self.model.parameters.set_values(
            self.res(),
            as_representation=True,
            parameters_identity=self.fit_parameters_identity,
        )
        if self.verbose > 1:
            AP_config.ap_logger.info(
                f"LM Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )

        return self

    def update_uncertainty(self):
        # set the uncertainty for each parameter
        cov = self.covariance_matrix
        if torch.all(torch.isfinite(cov)):
            try:
                self.model.parameters.set_uncertainty(
                    torch.sqrt(
                        torch.abs(torch.diag(cov))
                    ),
                    as_representation=False,
                    parameters_identity=self.fit_parameters_identity,
                )
            except RuntimeError as e:
                AP_config.ap_logger.warning(f"Unable to update uncertainty due to: {e}")

    @torch.no_grad()
    def undo_step(self) -> None:
        AP_config.ap_logger.info("undoing step, trying to recover")
        assert (
            self.decision_history.count("accept") >= 2
        ), "cannot undo with not enough accepted steps, retry with new parameters"
        assert len(self.decision_history) == len(self.lambda_history)
        assert len(self.decision_history) == len(self.L_history)
        found_accept = False
        for i in reversed(range(len(self.decision_history))):
            if not found_accept and self.decision_history[i] == "accept":
                found_accept = True
                continue
            if self.decision_history[i] != "accept":
                continue
            self.current_state = torch.tensor(
                self.lambda_history[i],
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
            self.L = self.L_history[i] * self.Lup

    def take_low_rho_step(self) -> bool:
        for i in reversed(range(len(self.decision_history))):
            if "accept" in self.decision_history[i]:
                return False
            if self.rho_history[i] is not None and self.rho_history[i] > 0:
                if self.verbose > 0:
                    AP_config.ap_logger.info(
                        f"taking a low rho step for some progress: {self.rho_history[i]}"
                    )
                self.current_state = torch.tensor(
                    self.lambda_history[i],
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
                self.L = self.L_history[i]

                self.loss_history.append(self.loss_history[i])
                self.L_history.append(self.L)
                self.lambda_history.append(
                    np.copy((self.current_state).detach().cpu().numpy())
                )
                self.decision_history.append("low rho accept")
                self.rho_history.append(self.rho_history[i])

                with torch.no_grad():
                    self.update_Yp(torch.zeros_like(self.current_state))
                    self.prev_Y[0] = self.prev_Y[1]
                    self.prev_Y[1] = self.current_Y
                self.update_J_AD()
                self.update_hess()
                self.update_grad(self.prev_Y[1])
                self.iteration += 1
                self.count_reject = 0
                return True

    @torch.no_grad()
    def update_h(self) -> torch.Tensor:
        """Solves the LM update linear equation (H + L*I)h = G to determine
        the proposal for how to adjust the parameters to decrease the
        chi2.

        """
        h = torch.zeros_like(self.current_state)
        if self.iteration == 0:
            return h

        h = torch.linalg.solve(
            (
                self.hess
                + self.L**2
                * torch.eye(
                    len(self.grad), dtype=AP_config.ap_dtype, device=AP_config.ap_device
                )
            )
            * (
                1
                + self.L**2
                * torch.eye(
                    len(self.grad), dtype=AP_config.ap_dtype, device=AP_config.ap_device
                )
            ) ** 2
            / (1 + self.L**2),
            self.grad,
        )
        return h

    @torch.no_grad()
    def update_Yp(self, h):
        """
        Updates the current model values for each pixel
        """
        # Sample model at proposed state
        self.current_Y = self.model(
            parameters=self.current_state + h,
            as_representation=True,
            parameters_identity=self.fit_parameters_identity,
            window=self.fit_window,
        ).flatten("data")

        # Add constraint evaluations
        if self.constraints is not None:
            for con in self.constraints:
                self.current_Y = torch.cat((self.current_Y, con(self.model)))

    @torch.no_grad()
    def update_chi2(self):
        """
        Updates the chi squared / ndf value
        """
        # Apply mask if needed
        if self.model.target.has_mask:
            loss = (
                torch.sum(
                    ((self.Y - self.current_Y) ** 2 * self.W)[
                        torch.logical_not(self.mask)
                    ]
                )
                / self.ndf
            )
        else:
            loss = torch.sum((self.Y - self.current_Y) ** 2 * self.W) / self.ndf

        return loss

    def update_J_AD(self) -> None:
        """
        Update the jacobian using automatic differentiation, produces an accurate jacobian at the current state.
        """
        # Free up memory
        del self.J
        if "cpu" not in AP_config.ap_device:
            torch.cuda.empty_cache()

        # Compute jacobian on image
        self.J = self.model.jacobian(
            torch.clone(self.current_state).detach(),
            as_representation=True,
            parameters_identity=self.fit_parameters_identity,
            window=self.fit_window,
        ).flatten("data")

        # compute the constraint jacobian if needed
        if self.constraints is not None:
            for con in self.constraints:
                self.J = torch.cat((self.J, con.jacobian(self.model)))

        # Apply mask if needed
        if self.model.target.has_mask:
            self.J[self.mask] = 0.0

        # Note that the most recent jacobian was a full autograd jacobian
        self.full_jac = True

    def update_J_natural(self) -> None:
        """
        Update the jacobian using automatic differentiation, produces an accurate jacobian at the current state. Use this method to get the jacobian in the parameter space instead of representation space.
        """
        # Free up memory
        del self.J
        if "cpu" not in AP_config.ap_device:
            torch.cuda.empty_cache()

        # Compute jacobian on image
        self.J = self.model.jacobian(
            torch.clone(
                self.model.parameters.transform(
                    self.current_state,
                    to_representation=False,
                    parameters_identity=self.fit_parameters_identity,
                )
            ).detach(),
            as_representation=False,
            parameters_identity=self.fit_parameters_identity,
            window=self.fit_window,
        ).flatten("data")

        # compute the constraint jacobian if needed
        if self.constraints is not None:
            for con in self.constraints:
                self.J = torch.cat((self.J, con.jacobian(self.model)))

        # Apply mask if needed
        if self.model.target.has_mask:
            self.J[self.mask] = 0.0

        # Note that the most recent jacobian was a full autograd jacobian
        self.full_jac = False

    @torch.no_grad()
    def update_J_Broyden(self, h, Yp, Yph) -> None:
        """
        Use the Broyden update to approximate the new Jacobian tensor at the current state. Less accurate, but far faster.
        """

        # Update the Jacobian
        self.J = Broyden_step(self.J, h, Yp, Yph)

        # Apply mask if needed
        if self.model.target.has_mask:
            self.J[self.mask] = 0.0

        # compute the constraint jacobian if needed
        if self.constraints is not None:
            for con in self.constraints:
                self.J = torch.cat((self.J, con.jacobian(self.model)))

        # Note that the most recent jacobian update was with Broyden step
        self.full_jac = False

    @torch.no_grad()
    def update_hess(self) -> None:
        """
        Update the Hessian using the jacobian most recently computed on the image.
        """

        if isinstance(self.W, float):
            self.hess = torch.matmul(self.J.T, self.J)
        else:
            self.hess = torch.matmul(self.J.T, self.W.view(len(self.W), -1) * self.J)
        self.hess += self.epsilon5 * torch.eye(
            len(self.current_state),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )

    @property
    @torch.no_grad()
    def covariance_matrix(self) -> torch.Tensor:
        if self._covariance_matrix is not None:
            return self._covariance_matrix
        self.update_J_natural()
        self.update_hess()
        try:
            self._covariance_matrix = 2*torch.linalg.inv(self.hess)
        except:
            AP_config.ap_logger.warning(
                "WARNING: Hessian is singular, likely at least one model is non-physical. Will massage Hessian to continue but results should be inspected."
            )
            self.hess += torch.eye(
                len(self.grad), dtype=AP_config.ap_dtype, device=AP_config.ap_device
            ) * (torch.diag(self.hess) == 0)
            self._covariance_matrix = 2*torch.linalg.inv(self.hess)
        return self._covariance_matrix

    @torch.no_grad()
    def update_grad(self, Yph) -> None:
        """
        Update the gradient using the model evaluation on all pixels
        """
        self.grad = torch.matmul(self.J.T, self.W * (self.Y - Yph))

    @torch.no_grad()
    def rho(self, Xp, Xph, h) -> torch.Tensor:
        return (
            self.ndf
            * (Xp - Xph)
            / abs(
                torch.dot(
                    h,
                    self.L**2 * (torch.abs(torch.diag(self.hess) - self.epsilon5) * h)
                    + self.grad,
                )
            )
        )

    def accept_history(self) -> (List[np.ndarray], List[np.ndarray], List[float]):
        lambdas = []
        Ls = []
        losses = []

        for l in range(len(self.decision_history)):
            if "accept" in self.decision_history[l] and np.isfinite(
                self.loss_history[l]
            ):
                lambdas.append(self.lambda_history[l])
                Ls.append(self.L_history[l])
                losses.append(self.loss_history[l])
        return lambdas, Ls, losses

    def progress_history(self) -> (List[np.ndarray], List[np.ndarray], List[float]):
        lambdas = []
        Ls = []
        losses = []

        for l in range(len(self.decision_history)):
            if self.decision_history[l] == "accept":
                lambdas.append(self.lambda_history[l])
                Ls.append(self.L_history[l])
                losses.append(self.loss_history[l])
        return lambdas, Ls, losses

      
        

class LM_Constraint:
    """Add an arbitrary constraint to the LM optimization algorithm.

    Expresses a constraint between parameters in the LM optimization
    routine. Constraints may be used to bias parameters to have
    certain behaviour, for example you may require the radius of one
    model to be larger than that of another, or may require two models
    to have the same position on the sky. The constraints defined in
    this object are fuzzy constraints and so can be broken to some
    degree, the amount of constraint breaking is determined my how
    informative the data is and how strong the constraint weight is
    set. To create a constraint, first construct a function which
    takes as argument a 1D tensor of the model parameters and gives as
    output a real number (or 1D tensor of real numbers) which is zero
    when the constraint is satisfied and non-zero increasing based on
    how much the constraint is violated. For example:

    def example_constraint(P):
        return (P[1] - P[0]) * (P[1] > P[0]).int()

    which enforces that parameter 1 is less than parameter 0. Note
    that we do not use any control flow "if" statements and instead
    incorporate the condition through multiplication, this is
    important as it allows pytorch to compute derivatives through the
    expression and performs far faster on GPU since no communication
    is needed back and forth to handle the if-statement. Keep this in
    mind while constructing your constraint function. Also, make sure
    that any math operations are performed by pytorch so it can
    construct a computational graph. Bayond the requirement that the
    constraint be differentiable, there is no limitation on what
    constraints can be built with this system.

    Args:
      constraint_func (Callable[torch.Tensor, torch.Tensor]): python function which takes in a 1D tensor of parameters and generates real values in a tensor.
      constraint_args (Optional[tuple]): An optional tuple of arguments for the constraint function that will be unpacked when calling the function.
      weight (torch.Tensor): The weight of this constraint in the range (0,inf). Smaller values mean a stronger constraint, larger values mean a weaker constraint. Default 1.
      representation_parameters (bool): if the constraint_func expects the parameters in the form of their representation or their standard value. Default False
      out_len (int): the length of the output tensor by constraint_func. Default 1
      reference_value (torch.Tensor): The value at which the constraint is satisfied. Default 0.
      reduce_ndf (float): Amount by which to reduce the degrees of freedom. Default 0.

    """

    def __init__(
        self,
        constraint_func: Callable[[torch.Tensor, Any], torch.Tensor],
        constraint_args: tuple = (),
        representation_parameters: bool = False,
        out_len: int = 1,
        reduce_ndf: float = 0.0,
        weight: Optional[torch.Tensor] = None,
        reference_value: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        self.constraint_func = constraint_func
        self.constraint_args = constraint_args
        self.representation_parameters = representation_parameters
        self.out_len = out_len
        self.reduce_ndf = reduce_ndf
        self.reference_value = torch.as_tensor(
            reference_value if reference_value is not None else torch.zeros(out_len),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )
        self.weight = torch.as_tensor(
            weight if weight is not None else torch.ones(out_len),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )

    def jacobian(self, model: "AstroPhot_Model"):
        jac = jacobian(
            lambda P: self.constraint_func(P, *self.constraint_args),
            model.parameters.get_vector(
                as_representation=self.representation_parameters
            ),
            strategy="forward-mode",
            vectorize=True,
            create_graph=False,
        )

        return jac.reshape(-1, np.sum(model.parameters.vector_len()))

    def __call__(self, model: "AstroPhot_Model"):
        return self.constraint_func(
            model.parameters.get_vector(
                as_representation=self.representation_parameters
            ),
            *self.constraint_args,
        )



import numpy as np
import torch

from matplotlib.patches import Ellipse, Rectangle, Polygon
from matplotlib import pyplot as plt
import matplotlib
from scipy.stats import iqr
from scipy.stats import norm

__all__ = ("covariance_matrix", )

def covariance_matrix(covariance_matrix, mean, labels = None, figsize = (10,10), reference_values = None, ellipse_colors='g', showticks = True, **kwargs):
    num_params = covariance_matrix.shape[0]
    fig, axes = plt.subplots(num_params, num_params, figsize=figsize)
    plt.subplots_adjust(wspace=0., hspace=0.)

    for i in range(num_params):
        for j in range(num_params):
            ax = axes[i, j]

            if i == j:
                x = np.linspace(mean[i] - 3 * np.sqrt(covariance_matrix[i, i]), mean[i] + 3 * np.sqrt(covariance_matrix[i, i]), 100)
                y = norm.pdf(x, mean[i], np.sqrt(covariance_matrix[i, i]))
                ax.plot(x, y, color='g')
                ax.set_xlim(mean[i] - 3 * np.sqrt(covariance_matrix[i, i]), mean[i] + 3 * np.sqrt(covariance_matrix[i, i]))
                if reference_values is not None:
                    ax.axvline(reference_values[i], color='red', linestyle='-', lw=1)
            elif j < i:
                cov = covariance_matrix[np.ix_([j, i], [j, i])]
                lambda_, v = np.linalg.eig(cov)
                lambda_ = np.sqrt(lambda_)
                angle = np.rad2deg(np.arctan2(v[1, 0], v[0, 0]))
                for k in [1, 2]:
                    ellipse = Ellipse(xy=(mean[j], mean[i]),
                                      width=lambda_[0] * k * 2,
                                      height=lambda_[1] * k * 2,
                                      angle=angle,
                                      edgecolor=ellipse_colors,
                                      facecolor='none')
                    ax.add_artist(ellipse)

                # Set axis limits
                margin = 3
                ax.set_xlim(mean[j] - margin * np.sqrt(covariance_matrix[j, j]), mean[j] + margin * np.sqrt(covariance_matrix[j, j]))
                ax.set_ylim(mean[i] - margin * np.sqrt(covariance_matrix[i, i]), mean[i] + margin * np.sqrt(covariance_matrix[i, i]))
                
                if reference_values is not None:
                    ax.axvline(reference_values[j], color='red', linestyle='-', lw=1)
                    ax.axhline(reference_values[i], color='red', linestyle='-', lw=1)
                
            if j > i:
                ax.axis('off')

            if i < num_params - 1:
                ax.set_xticklabels([])
            else:
                if labels is not None:
                    ax.set_xlabel(labels[j])
            if not showticks:
                ax.yaxis.set_major_locator(plt.NullLocator())

            if j > 0:
                ax.set_yticklabels([])
            else:
                if labels is not None:
                    ax.set_ylabel(labels[i])
            if not showticks:
                ax.xaxis.set_major_locator(plt.NullLocator())

    
    return fig, ax

if __name__ == "__main__":

    fig, ax = covariance_matrix(np.array([[4,-2], [-2,4]]), np.array([0,0]))
    plt.show()
